package vacworld;

import agent.State;
import java.io.*;

/** Represents a state in the vacuum world. */
public class VacuumState extends State {

  /* Constants for the initial state of the agent. */
  protected static int INIT_X = 1;
  protected static int INIT_Y = 1;
  protected static int INIT_DIR = Direction.SOUTH;

  private static final int CLEAR = 0;
  private static final int DIRT = 1;
  private static final int WALL = 2;
  
  /* Default map for initial state.
     Note the smaller arrays are columns in the map, not rows */
  protected static int[][] defaultMap = 
  {{WALL, WALL, WALL, WALL, WALL, WALL, WALL},
   {WALL, CLEAR, CLEAR, CLEAR, WALL, WALL, WALL},
   {WALL, CLEAR, DIRT, CLEAR, CLEAR, WALL, WALL},
   {WALL, CLEAR, CLEAR, CLEAR, CLEAR, CLEAR, WALL},
   {WALL, CLEAR, DIRT, CLEAR, CLEAR, DIRT, WALL},
   {WALL, WALL, WALL, WALL, CLEAR, CLEAR, WALL},
   {WALL, WALL, WALL, WALL, WALL, WALL, WALL}};

  protected static double DIRT_CHANCE = 0.3;
  protected static double WALL_CHANCE = 0.2;

  /* Variables for the state of the agent. */
  protected int agentX;
  protected int agentY;
  protected int agentDir;
  protected boolean robotOff = false;

  protected boolean bump = false;
  
  /** An array to that contains the locations of objects in the world. */
  protected int[][] map;

  protected int numDirtyLocs = 0;

  protected int height;
  protected int width;

  static final int DEFAULT_HEIGHT = 7;
  static final int DEFAULT_WIDTH = 7;

  /** Return the default initial state for the vacuum world. */
  public static VacuumState getInitState() {

    VacuumState state;

    state = new VacuumState();
    state.width = defaultMap.length;
    state.height = defaultMap[0].length;
    state.map = new int[state.width][state.height];
    state.map = defaultMap;
    state.numDirtyLocs = 2;
    state.agentX = INIT_X;
    state.agentY = INIT_Y;
    state.agentDir = INIT_DIR;
    return state;
  }

  /** Return a random initial state for the vacuum world. The agent's
      initial position is always the same, but the dirt locations and
      other obstacles are random. */
  public static VacuumState getRandomState() {

    VacuumState state;
    double roll;

    state = new VacuumState();
    state.numDirtyLocs = 0;
    state.height = DEFAULT_HEIGHT;
    state.width = DEFAULT_WIDTH;
    state.map = new int[state.width][state.height];
    for (int i=0; i < state.width; i++) {
      for (int j=0; j < state.height; j++) {
        if (i == 0 || i == (state.width - 1))
          state.map[i][j] = WALL;
        else {
          if (j == 0 || j == (state.height - 1))
            state.map[i][j] = WALL;
          else {
            roll = Math.random();
            if (i != INIT_X && i != INIT_Y && roll <= WALL_CHANCE) {
              state.map[i][j] = WALL;
            }
            else if (roll <= DIRT_CHANCE) {
              state.map[i][j] = DIRT;
              state.numDirtyLocs++;
            }
            else
              state.map[i][j] = CLEAR;
          }
        }
      }
    }   
    state.agentX = INIT_X;
    state.agentY = INIT_Y;
    state.agentDir = INIT_DIR;
    return state;
  }

  /** Construct a new vacuum state. */
  public VacuumState() {
  }

  /** Return the agent's X position. */
  public int getAgentX() {
    return agentX;
  }

  /** Return the agent's Y position. */
  public int getAgentY() {
    return agentY;
  }

  /** Return the agent's orientation. Note that the constants in
      the Direction class are used to interpret the meaning of this
      result.*/
  public int getAgentDir() {
    return agentDir;
  }

  /** Return a string the represents the agent's orientation. */
  public String getAgentDirString() {
    return Direction.toString(agentDir);
  }
  
  /** Change the agent's X position. */
  public void setAgentX(int x) {
    agentX = x;
  }

  /** Change the agent's Y position. */
  public void setAgentY(int y) {
    agentY = y;
  }

  /** Change the agent's orientation. */
  public void setAgentDir(int dir) {
    agentDir = dir;
  }

  /** Set an indicator that the robot has shut down in the state. */
  public void setRobotOff(boolean off) {
    robotOff = off;
  }

  /** Set an indicator that the robot bumped into an obstacle 
      during its previous move. */
  public void setBump(boolean bump) {
    this.bump = bump;
  }

  /** Remove dirt from the specified location. */
  public void removeDirt(int x, int y) {

    if (map[x][y] == DIRT) {
      map[x][y] = CLEAR;
      numDirtyLocs--;
    }
  }

  /** Return true if the specified location has dirt in it. */
  public boolean hasDirt(int x, int y) {

    if (map[x][y] == DIRT)
      return true;
    else
      return false;
  }

  /** Return true if the specified location has an obstacle in it. */
  public boolean hasObstacle(int x, int y) {

    if (map[x][y] == WALL)
      return true;
    else
      return false;
  }

  /** Return true if the agent bumped into an obstacle on its previous
      move. */
  public boolean bumped() {

    return bump;
  }

  /** Return true if the location is within bounds of the state's map. */
  public boolean inBounds(int x, int y) {

    if (x >= 0 && x < width && y >= 0 && y < height)
      return true;
    else
      return false;
  }

  /** Return true if the robot is turned off in the state. */
  public boolean isRobotOff() {

    return robotOff;
  }

  /** Return the number of dirty locations in the state. */
  public int getNumDirtyLocs() {

    return numDirtyLocs;
  }

  /** Print an output of the state to the screen. This output includes
      a map as well as information about the agent's location and the
      direction it is facing. On the map, "A" denotes the agent, "*"
      denotes dirt, and "X" denotes an obstacle. */
  public void display() {

    String input;

    for (int j=0; j < width; j++)
      System.out.print("  " + j);
    System.out.println();

    System.out.print(" ");
    for (int j=0; j < width; j++)
      System.out.print("+--");
    System.out.println("+");

    /* To print to the screen properly, the loop in the Y
       direction must be the outer loop. */
    for (int i=0; i < height; i++) {
      System.out.print(i + "|");
      for (int j=0; j < width; j++) {
        if (i == agentY && j == agentX)
          System.out.print("A");
        else if (hasDirt(j,i))
          System.out.print("*");
        else if (hasObstacle(j,i))
          System.out.print("X");
        else
          System.out.print(" ");
        System.out.print(" |");
      }
      System.out.println();

      System.out.print(" +");
      for (int j=0; j < width - 1; j++)
        System.out.print("--+");
      System.out.println("--+");

    }
    System.out.println("Location: (" + agentX + "," + agentY + ")  " +
                       "Facing: " + getAgentDirString());
    System.out.println();
    
    System.out.println("Press RETURN to continue.");
    
    BufferedReader console = new BufferedReader(new InputStreamReader(System.in));
    try {
      input = console.readLine();
    }
    catch (IOException e) {
      System.out.println(e.getMessage());
      return;
    }

  }


}



